import logging
import re

import discord
from discord.ext import commands

from .permissions import check_missing_channel_permissions, find_first_accessible_text_channel


def extract_invite_code(invite_url: str) -> str | None :
	"""Extracts the invite code from a Discord invite URL."""
	match = re.search(r"(?:https?://)?(?:www\.)?discord(?:\.gg|app\.com/invite)/([a-zA-Z0-9-]+)", invite_url)
	return match.group(1) if match else invite_url


async def check_guild_invites(bot: commands.AutoShardedBot, guild: discord.Guild, current_invite=None, channel: discord.TextChannel | discord.ForumChannel | discord.VoiceChannel = None) -> str :
	invite: None | discord.Invite = None

	if current_invite :
		try :
			invite_code = extract_invite_code(current_invite)
			invite = await bot.fetch_invite(invite_code)
			return invite.url
		except discord.HTTPException or discord.NotFound :
			logging.info(f"{guild.name}'s invite expired, creating a new one.")
		except ValueError:
			logging.info(f"{guild.name}'s has weird characters in invite code, {current_invite}")
	try :
		if not channel:
			channel = find_first_accessible_text_channel(guild)
		if len(check_missing_channel_permissions(channel, ["view_channel", "create_instant_invite"])) > 0 :
			invite: str = await create_invite(channel, reason=f"Missing Invite/Invalid invite, making a new one.")
		return invite
	except discord.Forbidden :
		logging.info(f"No permission to create invites in {guild.name}")
	try :
		invite: discord.Invite = (await guild.invites())[0]
		return invite.url
	except discord.Forbidden :
		logging.info(f"No permission to fetch invites in {guild.name}")
		return "forbidden"
	except IndexError :
		logging.info(f"{guild.name} has no invites.")
		return "no invites available"
	except Exception as e :
		logging.error(f"Error creating invite: {e}")
		return "error creating invite"


async def create_invite(channel: discord.TextChannel, reason=None, max_age=604800) -> str :
	try :
		invite = await channel.create_invite(max_age=604800, reason=reason)
		return invite.url
	except discord.Forbidden :
		return 'No permission'
	except Exception as e :
		logging.warning(f"Error creating invite: {e}")
		return f'No permission/Error'



import asyncio
import logging

import discord
from discord import CategoryChannel, DMChannel, ForumChannel, GroupChannel, Member, StageChannel, TextChannel, Thread, \
	User, \
	VoiceChannel
from discord.ext.commands.help import MISSING

from .exceptions import NoChannelException, NoPermissionException
from .permissions import check_missing_channel_permissions

MAX_LENGTH = 1800


async def handle_no_permission(channel: User | Member | VoiceChannel | StageChannel | TextChannel | ForumChannel | CategoryChannel | Thread | DMChannel | GroupChannel | None,
                               error_mode: str = 'error') :
	"""
	Handles no permission errors within the code based on error_modes: error, warn, ignore.
	Error: Raises NoPermissionException
	Warn: Logs a warning message and sends a message to the owner of the guild.
	Ignore: Does nothing.
	:param channel:
	:param error_mode:
	"""
	match error_mode.lower() :
		case 'error' :
			missing_perms = check_missing_channel_permissions(channel,
			                                                  ['view_channel', 'send_messages', 'embed_links',
			                                                   'attach_files'])
			raise NoPermissionException(required_perms=missing_perms, channel=channel)
		case 'warn' :
			missing_perms = check_missing_channel_permissions(channel,
			                                                  ['view_channel', 'send_messages', 'embed_links',
			                                                   'attach_files'])
			logging.warning(f"Missing permission to send message to {channel.name}")
			await channel.guild.owner.send(
				f"Missing permission to send message to {channel.name}. Check permissions: {', '.join(missing_perms)}", )
		case 'ignore' :
			logging.debug(f"Missing permission to send message to {channel.name}. Ignoring.")
		case _ :
			logging.debug(f"Missing valid error_mode, defaulting to error.")
			missing_perms = check_missing_channel_permissions(channel,
			                                                  ['view_channel', 'send_messages', 'embed_links', 'attach_files'])
			raise NoPermissionException(required_perms=missing_perms, channel=channel)


async def send_message(channel: User | Member | VoiceChannel | StageChannel | TextChannel | ForumChannel | CategoryChannel | Thread | DMChannel | GroupChannel | None, message: str = None, embed=None,
                       view=None, files=None, error_mode: str = 'error') -> discord.Message :
	"""Send a message to a channel, if there is no permission it will send an error message to the owner. Automatically handles messages longer than 2000 characters by splitting them into multiple messages.
	:param channel:
	:param message:
	:param embed:
	:param view:
	:param files:
	:param error_mode:
	:return:
	"""
	last_message = None
	if channel is None :
		raise NoChannelException
	try :
		length = 0
		if not message or len(message) < 1 :
			message = " "
		while length < len(message) :
			last_message = await channel.send(message[length :length + MAX_LENGTH], embed=embed, view=view, files=files)
			length += MAX_LENGTH
		else :
			return last_message
	except discord.errors.Forbidden :
		await handle_no_permission(channel, error_mode)


async def send_response(interaction: discord.Interaction, response, ephemeral=False, view=MISSING, embed=MISSING, error_mode = 'error') :
	"""Sends a response to an interaction. If the interaction is already responded, it will send a followup message.
	:param error_mode:
	:param interaction:
	:param response:
	:param ephemeral:
	:param view:
	:param embed:
	:return:
	"""
	try :
		if not response or len(response) < 1 :
			response = " "
		return await interaction.response.send_message(response, ephemeral=ephemeral, view=view, embed=embed)
	except discord.errors.Forbidden :
		await handle_no_permission(interaction.channel, error_mode=error_mode)
	except discord.errors.NotFound or discord.InteractionResponded :
		try :
			await interaction.followup.send(
				response,
				ephemeral=ephemeral,
				view=view,
				embed=embed,

			)
		except discord.errors.NotFound :
			await send_message(interaction.channel, response, view=view, embed=embed)


	except Exception :
		return await send_message(interaction.channel ,response, view=view, embed=embed)


async def await_message(interaction, message) -> discord.Message | bool :
	"""Wait for a message from the user that triggered the interaction. If the message is 'cancel', return False.
	:param interaction:
	:param message:
	:return:
	"""
	msg: discord.Message = await send_message(interaction.channel,
	                                          message)

	def check_message(m, interaction) :
		return m.author == interaction.user and m.channel == interaction.channel

	m = await interaction.client.wait_for('message', check=lambda m : check_message(m, interaction), timeout=600)
	await msg.delete()
	if m.content.lower() == "cancel" :
		return False
	return m


async def fetch_message_or_none(channel: discord.TextChannel | discord.DMChannel | discord.Thread,
                                id: int, wait = 0) -> discord.Message | None :
	"""
	Fetch a message from a channel. If the message is not found, return None. Wait for a specified amount of time before fetching the message; useful for fetching starting message of threads as they often have delays.
	:param wait:
	:return:
	:param channel:
	:param id:
	:return:
	"""
	try :
		await asyncio.sleep(wait)
		return await channel.fetch_message(id)
	except discord.errors.NotFound :
		return None


async def delete_message(message: discord.Message | discord.Thread) :
	"""
	Delete a message. If the message is not found, return None. If the message is not found, send a message to the channel.
	:param message:
	:return:
	"""
	try :
		await message.delete()
	except discord.errors.Forbidden :
		if isinstance(message, discord.Message) :
			await send_message(message.channel, f"Could not delete message {message.jump_url}")
			return
		await send_message(message, f"Could not delete message {message.jump_url}")
	except Exception as e :
		logging.error(
			f"Failed to delete {message} because {e}")




import logging

import discord


def create_channel(guild: discord.Guild, channel_name: str,
                   channel_type: discord.ChannelType = discord.ChannelType.text,
                   category: discord.CategoryChannel = None) :
	"""
	Create a channel in a given guild.
	:return:
	:param guild: The guild (server) where the channel will be created.
	:param channel_name: The name of the channel to be created.
	:param channel_type: The type of the channel (text or voice).
	:param category: The category under which the channel will be created (optional).
	:return: The created channel object.
	"""
	match channel_type :
		case discord.ChannelType.voice :
			# Creates voice channels
			return guild.create_voice_channel(name=channel_name, category=category)
		case discord.ChannelType.forum :
			# Creates forum channels
			return guild.create_forum(name=channel_name, category=category)
		case discord.ChannelType.stage_voice :
			# Creates stage channels
			return guild.create_stage_channel(name=channel_name, category=category)
		case _ :
			# Creates text channels and handles undefined channel types
			return guild.create_text_channel(name=channel_name, category=category)

def delete_channel(channel: discord.abc.GuildChannel, reason = "Deleted by bot") :
	"""
	Delete a channel.
	:param reason:
	:param channel: The channel to be deleted.
	:return: None
	"""
	try:
		return channel.delete(reason=reason)
	except discord.Forbidden:
		logging.error(f"Failed to delete channel {channel.name}({channel.id}): FORBIDDEN")
		return None


